#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stddef.h>

#include "TutMath.h"
#include <GL\glew.h>

#define WINDOW_WIDTH					1024
#define WINDOW_HEIGHT					768

#define COUNT_OF(x) (sizeof(x) / sizeof(x[0]))

static bool running = true;

char* vertex_shader_str =
"#version 400 core\n\
layout(location = 0) in vec4 position; \
void main() { \
	gl_Position = position; \
}";

char* frag_shader_str =
"#version 400 core\n\
layout(location = 0) out vec4 outColor;\n\
void main() {\
	outColor = vec4(1, 1, 1, 1); \
}";

bool CheckShaderCompileStatus(GLint shader)
{
	GLint status;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
	if (status == GL_FALSE)
	{
		char errorBuf[1024];
		glGetShaderInfoLog(shader, 1024, NULL, errorBuf);
		OutputDebugStringA(errorBuf);
	}
	return status == GL_TRUE;
}

bool CompileShaders(GLint *compiledProgram, char *vertexShader, char *fragmentShader)
{
	bool success = true;
	GLuint vert_shader = glCreateShader(GL_VERTEX_SHADER);
	GLuint frag_shader = glCreateShader(GL_FRAGMENT_SHADER);
	int vert_shader_len = strlen(vertexShader);
	int frag_shader_len = strlen(fragmentShader);
	glShaderSource(vert_shader, 1, &vertexShader, &vert_shader_len);
	glShaderSource(frag_shader, 1, &fragmentShader, &frag_shader_len);
	glCompileShader(vert_shader);
	glCompileShader(frag_shader);
	if (!CheckShaderCompileStatus(vert_shader) || !CheckShaderCompileStatus(frag_shader))
	{
		glDeleteShader(vert_shader);
		glDeleteShader(frag_shader);
		return false;
	}

	GLuint program = glCreateProgram();
	glAttachShader(program, vert_shader);
	glAttachShader(program, frag_shader);
	glLinkProgram(program);
	GLint status;
	glGetProgramiv(program, GL_LINK_STATUS, &status);
	if (status == GL_FALSE)
	{
		char errorBuf[1024];
		glGetProgramInfoLog(program, 1024, NULL, errorBuf);
		OutputDebugStringA(errorBuf);
		success = false;
	}
	else
	{
		*compiledProgram = program;
	}
	glDetachShader(program, vert_shader);
	glDetachShader(program, frag_shader);
	glDeleteShader(vert_shader);
	glDeleteShader(frag_shader);
	return success;
}

LONG WINAPI WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_CLOSE:
	case WM_QUIT:
		running = false;
		break;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}

int CALLBACK WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	LPWSTR wc_name = L"OGLWindowClass";

	WNDCLASS wc = {};
	wc.style = CS_OWNDC;
	wc.lpfnWndProc = (WNDPROC)WindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(NULL, IDI_WINLOGO);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = wc_name;

	if (!RegisterClass(&wc)) {
		MessageBoxA(NULL, "Failed to register window class!", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}

	// make sure the client area is of specified size, not the window
	RECT rect;
	rect.right = WINDOW_WIDTH;
	rect.bottom = WINDOW_HEIGHT;
	rect.left = 0;
	rect.top = 0;
	if (!AdjustWindowRectEx(&rect, WS_BORDER | WS_CAPTION, false, 0))
	{
		MessageBoxA(NULL, "Adjusting size failed", "Error", MB_OK | MB_ICONERROR);
	}

	HWND hWnd = CreateWindowEx(
		0, // Optional window styles.
		wc_name, // Window class
		L"OpenGL Tutorial",	// Window text
		WS_OVERLAPPEDWINDOW, // Window style
		CW_USEDEFAULT, CW_USEDEFAULT, rect.right - rect.left, rect.bottom - rect.top, // position and size
		NULL, // Parent window
		NULL, // Menu
		hInstance,// Instance handle
		NULL // Additional application data
	);
	if (hWnd == NULL) {
		MessageBoxA(NULL, "Failed to create window", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}

	HDC hdc = GetDC(hWnd);
	if (hdc == NULL)
	{
		MessageBoxA(NULL, "Getting device context failed", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}

	// Choose pixel format
	PIXELFORMATDESCRIPTOR pixel_format_desc = {};
	pixel_format_desc.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pixel_format_desc.nVersion = 1;
	pixel_format_desc.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pixel_format_desc.iPixelType = PFD_TYPE_RGBA;
	pixel_format_desc.cColorBits = 32;
	pixel_format_desc.cDepthBits = 24;
	int pixel_format = ChoosePixelFormat(hdc, &pixel_format_desc);
	if (pixel_format <= 0)
	{
		MessageBoxA(NULL, "Choosing pixel format failed", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}

	if (DescribePixelFormat(hdc, pixel_format, sizeof(PIXELFORMATDESCRIPTOR), &pixel_format_desc) == 0)
	{
		MessageBoxA(NULL, "DescibePixelFormat failed", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}

	if (SetPixelFormat(hdc, pixel_format, &pixel_format_desc) == FALSE)
	{
		MessageBoxA(NULL, "Setting pixel format failed", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}

	HGLRC gl_ctx = wglCreateContext(hdc);
	if (gl_ctx == NULL)
	{
		MessageBoxA(NULL, "Creating OpenGL context failed", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}
	if (wglMakeCurrent(hdc, gl_ctx) == NULL)
	{
		MessageBoxA(NULL, "wglMakeCurrent failed", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}
	glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);

	glewInit();
	if (!glewIsSupported("GL_VERSION_4_0"))
	{
		MessageBoxA(hWnd, "Getting OpenGL extensions failed\n OpenGL 4.0+ is required", "Error", MB_OK | MB_ICONERROR);
		return -1;
	}

	GLint defaultProgram;
	if (!CompileShaders(&defaultProgram, vertex_shader_str, frag_shader_str))
		return -1;


	ShowWindow(hWnd, nCmdShow);

	MSG msg;
	while (running) // main loop
	{
		while (PeekMessage(&msg, hWnd, 0, 0, 1))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		glColorMask(1, 1, 1, 1);
		glClearColor(1.0f, 0.0f, 0.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glUseProgram(defaultProgram);

		SwapBuffers(hdc);
	}

	// cleanup
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(gl_ctx);
	ReleaseDC(hWnd, hdc);

	return 0;
}